# -*- coding: utf-8 -*-
# Copyright 2017 Interstellar Technologies Inc. All Rights Reserved.
# 東京−大阪間600kmを重力のみで最速で到達する経路を見つける
# 摩擦などはないと仮定している。
# スケールはDIDOのマニュアル参考

from __future__ import print_function
import numpy as np
import matplotlib.pyplot as plt
from OpenGoddard.optimize import Problem, Guess, Condition, Dynamics


class Ball:
    def __init__(self):
        self.g = 9.8  # gravity [m/s2]
        self.l = 600000  # goal [m]
        self.h = 300000  # depth limit [m]


def dynamics(prob, obj, section):
    x = prob.states(0, section)
    y = prob.states(1, section)
    v = prob.states(2, section)
    theta = prob.controls(0, section)

    g = obj.g

    dx = Dynamics(prob, section)
    dx[0] = v * np.sin(theta)
    dx[1] = v * np.cos(theta)
    dx[2] = g * np.cos(theta)
    return dx()


def equality(prob, obj):
    x = prob.states_all_section(0)
    y = prob.states_all_section(1)
    v = prob.states_all_section(2)
    theta = prob.controls_all_section(0)
    tf = prob.time_final(-1)

    result = Condition()

    # event condition
    result.add(x[0] - 0.0)
    result.add(y[0] - 0.0)
    result.add(v[0] - 0.0)
    result.add(x[-1] - obj.l)
    result.add(y[-1] - 0.0)
    # result.add(v[-1] - 0.0)

    return result()


def inequality(prob, obj):
    x = prob.states_all_section(0)
    y = prob.states_all_section(1)
    v = prob.states_all_section(2)
    theta = prob.controls_all_section(0)
    tf = prob.time_final(-1)

    result = Condition()
    # lower bounds
    # result.add(tf - 500)
    result.add(x - 0)
    result.add(y - 0)
    result.add(theta - 0)
    # result.add(100000 / obj.unit_y - y)
    # upper bounds
    result.add(np.pi - theta)
    result.add(obj.l - x)
    # result.add(700 - tf)

    return result()


def cost(prob, obj):
    tf = prob.time_final(-1)
    return tf


def cost_derivative(prob, obj):
    jac = Condition(prob.number_of_variables)
    index_tf = prob.index_time_final(0)
    # index_tf = prob.index_time_final(-1)
    jac.change_value(index_tf, 1)
    return jac()

# ========================
plt.close("all")
plt.ion()
# Program Starting Point
time_init = [0.0, 600.0]
n = [30]
num_states = [3]
num_controls = [1]
max_iteration = 10

flag_savefig = True

savefig_dir = "01_Brachistochrone/Tokyo_"

# ------------------------
# set OpenGoddard class for algorithm determination
prob = Problem(time_init, n, num_states, num_controls, max_iteration)
obj = Ball()
# obj.make_scales()

unit_x = 300000
unit_y = 100000
unit_t = 100
unit_v = unit_x / unit_t
prob.set_unit_states_all_section(0, unit_x)
prob.set_unit_states_all_section(1, unit_y)
prob.set_unit_states_all_section(2, unit_v)
prob.set_unit_controls_all_section(0, 1.0)
prob.set_unit_time(unit_t)

# ========================
# Initial parameter guess
half_nodes = int(prob.nodes[0] / 2)
theta_init = Guess.linear(prob.time_all_section, 0.0, np.pi)
# Guess.plot(prob.time_all_section, theta_init, "gamma", "time", "gamma")
# if(flag_savefig):plt.savefig(savefig_dir + "guess_gamma" + ".png")

x_init = Guess.linear(prob.time_all_section, 0.0, obj.l)
# Guess.plot(prob.time_all_section, x_init, "x", "time", "x")
# if(flag_savefig):plt.savefig(savefig_dir + "guess_x" + ".png")

# y_init = Guess.linear(prob.time_all_section, 0.0, obj.l / np.sqrt(3))
y_init0 = Guess.linear(prob.time_all_section[:half_nodes], 0, obj.h)
y_init1 = Guess.linear(prob.time_all_section[half_nodes:], obj.h, 0)
y_init = np.hstack((y_init0, y_init1))
# Guess.plot(prob.time_all_section, y_init, "y", "time", "y")
# if(flag_savefig):plt.savefig(savefig_dir + "guess_y" + ".png")

v_init0 = Guess.linear(prob.time_all_section[:half_nodes], 0, obj.h)
v_init1 = Guess.linear(prob.time_all_section[half_nodes:], obj.h, 0)
v_init = np.hstack((v_init0, v_init1))
# Guess.plot(prob.time_all_section, v_init, "v", "time", "v")
# if(flag_savefig):plt.savefig(savefig_dir + "guess_v" + ".png")


prob.set_states_all_section(0, x_init)
prob.set_states_all_section(1, y_init)
prob.set_controls_all_section(0, theta_init)

# ========================
# Main Process
# Assign problem to SQP solver
prob.dynamics = [dynamics]
prob.knot_states_smooth = []
prob.cost = cost
prob.cost_derivative = cost_derivative
prob.equality = equality
prob.inequality = inequality


def display_func():
    tf = prob.time_final(-1)
    print("tf: {0:.5f}".format(tf))


prob.solve(obj, display_func)

# ========================
# Post Process
# ------------------------
# Convert parameter vector to variable
x = prob.states_all_section(0)
y = prob.states_all_section(1)
v = prob.states_all_section(2)
theta = prob.controls_all_section(0)
time = prob.time_update()

# ------------------------
# Visualizetion
plt.figure()
plt.subplot(2, 1, 1)
plt.plot(time, x/1000, marker="o", label="x")
plt.plot(time, y/1000, marker="o", label="y")
plt.plot(time, v, marker="o", label="v")
for line in prob.time_knots():
    plt.axvline(line, color="k", alpha=0.5)
plt.grid()
plt.ylabel("position [km], velocity [m/s]")
plt.legend(loc="best")

plt.subplot(2, 1, 2)
plt.plot(time, theta, marker="o", label="gamma")
for line in prob.time_knots():
    plt.axvline(line, color="k", alpha=0.5)
plt.grid()
plt.xlabel("time [s]")
plt.ylabel("angle [rad]")
plt.legend(loc="best")
if(flag_savefig): plt.savefig(savefig_dir + "plot" + ".png")

plt.figure()
plt.plot(x/1000, y/1000, marker="o", label="trajectry")
plt.axhline(0, color="k")
plt.axvline(0, color="k")
plt.axvline(obj.l/1000, color="k")
plt.grid()
plt.xlabel("x [km]")
plt.ylabel("y [km]")
plt.axis('equal')
plt.legend(loc="best")
plt.gca().invert_yaxis()
if(flag_savefig): plt.savefig(savefig_dir + "trajectry" + ".png")

plt.show()



# -*- coding: utf-8 -*-
# Copyright 2017 Interstellar Technologies Inc. All Rights Reserved.

from __future__ import print_function
import numpy as np
import matplotlib.pyplot as plt
from OpenGoddard.optimize import Problem, Guess, Condition, Dynamics


class Orbiter:
    def __init__(self):
        self.u_max = 0.01
        self.r0 = 1.0
        self.vr0 = 0.0
        self.vt0 = 1.0
        self.rf = 4.0
        self.vrf = 0.0
        self.vtf = 0.5
        self.tf_max = 55


def dynamics(prob, obj, section):
    r   = prob.states(0, section)
    vr  = prob.states(1, section)
    vt  = prob.states(2, section)
    ur1 = prob.controls(0, section)
    ur2 = prob.controls(1, section)
    ut1 = prob.controls(2, section)
    ut2 = prob.controls(3, section)

    dx = Dynamics(prob, section)
    dx[0] = vr
    dx[1] = vt**2 / r - 1 / r**2 + (ur1 - ur2)
    dx[2] = - vr * vt / r + (ut1 - ut2)
    return dx()


def equality(prob, obj):
    r   = prob.states_all_section(0)
    vr  = prob.states_all_section(1)
    vt  = prob.states_all_section(2)
    ur1 = prob.controls_all_section(0)
    ur2 = prob.controls_all_section(1)
    ut1 = prob.controls_all_section(2)
    ut2 = prob.controls_all_section(3)
    tf  = prob.time_final(-1)

    result = Condition()

    # event condition
    result.add(r[0] - obj.r0)
    result.add(vr[0] - obj.vr0)
    result.add(vt[0] - obj.vt0)
    result.add(r[-1] - obj.rf)
    result.add(vr[-1] - obj.vrf)
    result.add(vt[-1] - obj.vtf)

    return result()


def inequality(prob, obj):
    r   = prob.states_all_section(0)
    vr  = prob.states_all_section(1)
    vt  = prob.states_all_section(2)
    ur1 = prob.controls_all_section(0)
    ur2 = prob.controls_all_section(1)
    ut1 = prob.controls_all_section(2)
    ut2 = prob.controls_all_section(3)
    tf  = prob.time_final(-1)

    result = Condition()

    # lower bounds
    result.add(r - obj.r0)
    result.add(ur1 - 0.0)
    result.add(ut1 - 0.0)
    result.add(ur2 - 0.0)
    result.add(ut2 - 0.0)
    result.add(tf - 0.0)

    # upper bounds
    result.add(obj.rf - r)
    result.add(obj.u_max - ur1)
    result.add(obj.u_max - ut1)
    result.add(obj.u_max - ur2)
    result.add(obj.u_max - ut2)
    result.add(obj.tf_max - tf)

    return result()


def cost(prob, obj):
    return 0.0


def running_cost(prob, obj):
    ur1 = prob.controls_all_section(0)
    ur2 = prob.controls_all_section(1)
    ut1 = prob.controls_all_section(2)
    ut2 = prob.controls_all_section(3)

    return (ur1 + ur2) + (ut1 + ut2)


# ========================
plt.close("all")
plt.ion()
# Program Starting Point
time_init = [0.0, 10.0]
n = [100]
num_states = [3]
num_controls = [4]
max_iteration = 10

flag_savefig = True

savefig_dir = "10_Low_Thrust_Orbit_Transfer/"

# ------------------------
# set OpenGoddard class for algorithm determination
prob = Problem(time_init, n, num_states, num_controls, max_iteration)
obj = Orbiter()

# ========================
# Initial parameter guess
r_init = Guess.linear(prob.time_all_section, obj.r0, obj.rf)
# Guess.plot(prob.time_all_section, r_init, "r", "time", "r")
# if(flag_savefig):plt.savefig(savefig_dir + "guess_r" + savefig_add + ".png")

vr_init = Guess.linear(prob.time_all_section, obj.vr0, obj.vrf)
# Guess.plot(prob.time_all_section, vr_init, "vr", "time", "vr")
# if(flag_savefig):plt.savefig(savefig_dir + "guess_vr" + savefig_add + ".png")

vt_init = Guess.linear(prob.time_all_section, obj.vt0, obj.vtf)
# Guess.plot(prob.time_all_section, theta_init, "vt", "time", "vt")
# if(flag_savefig):plt.savefig(savefig_dir + "guess_vt" + savefig_add + ".png")

ur1_init = Guess.linear(prob.time_all_section, obj.u_max, obj.u_max)
# Guess.plot(prob.time_all_section, ur1_init, "ur1", "time", "ur1")
# if(flag_savefig):plt.savefig(savefig_dir + "guess_ur1" + savefig_add + ".png")

ut1_init = Guess.linear(prob.time_all_section, obj.u_max, obj.u_max)
# Guess.plot(prob.time_all_section, ut1_init, "ut1", "time", "ut1")
# if(flag_savefig):plt.savefig(savefig_dir + "guess_ut1" + savefig_add + ".png")

prob.set_states_all_section(0, r_init)
prob.set_states_all_section(1, vr_init)
prob.set_states_all_section(2, vt_init)
prob.set_controls_all_section(0, ur1_init)
prob.set_controls_all_section(2, ut1_init)

# ========================
# Main Process
# Assign problem to SQP solver
prob.dynamics = [dynamics]
prob.knot_states_smooth = []
prob.cost = cost
prob.running_cost = running_cost
prob.equality = equality
prob.inequality = inequality


def display_func():
    tf = prob.time_final(-1)
    print("tf: {0:.5f}".format(tf))


prob.solve(obj, display_func, ftol=1e-12)

# ========================
# Post Process
# ------------------------
# Convert parameter vector to variable
r = prob.states_all_section(0)
vr = prob.states_all_section(1)
vt = prob.states_all_section(2)
ur1 = prob.controls_all_section(0)
ur2 = prob.controls_all_section(1)
ut1 = prob.controls_all_section(2)
ut2 = prob.controls_all_section(3)
time = prob.time_update()

# ------------------------
# Visualizetion
plt.figure()
plt.plot(time, r, marker="o", label="r")
for line in prob.time_knots():
    plt.axvline(line, color="k", alpha=0.5)
plt.grid()
plt.xlabel("time [-]")
plt.ylabel("r [-]")
plt.legend(loc="best")
if(flag_savefig): plt.savefig(savefig_dir + "r" + ".png")

plt.figure()
plt.plot(time, vr, marker="o", label="vr")
plt.plot(time, vt, marker="o", label="vt")
for line in prob.time_knots():
    plt.axvline(line, color="k", alpha=0.5)
plt.grid()
plt.xlabel("time [-]")
plt.ylabel("velocity [-]")
plt.legend(loc="best")
if(flag_savefig): plt.savefig(savefig_dir + "velocity" + ".png")

plt.figure()
plt.plot(time, (ur1 - ur2), marker="o", label="ur")
plt.plot(time, (ut1 - ut2), marker="o", label="ut")
# plt.plot(time, ur1, marker="o", label="ur1")
# plt.plot(time, ur2, marker="o", label="ur2")
# plt.plot(time, ut1, marker="o", label="ut1")
# plt.plot(time, ut2, marker="o", label="ut2")
plt.grid()
plt.xlabel("time [-]")
plt.ylabel("thrust [-]")
# plt.ylim([-0.02, 0.6])
plt.legend(loc="best")
if(flag_savefig): plt.savefig(savefig_dir + "thrust" + ".png")


from scipy import integrate
from scipy import interpolate
theta = integrate.cumtrapz(vt / r, time, initial=0)
theta_f = interpolate.interp1d(time, theta)
r_f = interpolate.interp1d(time, r)
time_fine = np.linspace(time[0], time[-1], 1000)
r_fine = r_f(time_fine)
theta_fine = theta_f(time_fine)
fig = plt.figure()
# plt.plot(r*np.cos(theta), r*np.sin(theta))
plt.plot(r_fine*np.cos(theta_fine), r_fine*np.sin(theta_fine))
ax = fig.add_subplot(111)
circle0 = plt.Circle((0.0, 0.0), 1.0, ls="--", fill=False, fc='none')
circlef = plt.Circle((0.0, 0.0), 4.0, ls="--", fill=False, fc='none')
ax.add_patch(circle0)
ax.add_patch(circlef)
plt.grid()
plt.axis('equal')
plt.ylim((-4.1, 4.1))
if(flag_savefig): plt.savefig(savefig_dir + "trajectry" + ".png")


plt.show()



# -*- coding: utf-8 -*-
# Copyright (c) 2017 Interstellar Technologies Inc. All Rights Reserved.
# Authors : Takahiro Inagawa, Kazuki Sakaki
#
# Lisence : MIT Lisence
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

"""
OpenGoddard.optimize - Optimal Trajectories module with PseudoSpectral Method
"""


import numpy as np
from scipy import special
from scipy import interpolate
from scipy import optimize
import matplotlib.pyplot as plt


class Problem:
    """ OpenGoddard Problem class.

    Args:
        time_init (list of float) : [time_start, time_section0, time_section0, , , time_final]
        nodes (int) : number of nodes
        number_of_states (list) : number of states
        number_of_controls (list) : number of controls
        maxIterator (int) : iteration max

    Attributes:
        nodes (int) : time nodes.
        number_of_states (int) : number of states.
        number_of_controls (int) : number of controls
        number_of_section (int) : number of section
        number_of_param (int) : number of inner variables
        div (list) : division point of inner variables
        tau : Gauss nodes
        w : weights of Gaussian quadrature
        D :  differentiation matrix of Gaussian quadrature
        time : time
        maxIterator (int) : max iterator
        time_all_section : all section time
        unit_states (list of float) : canonical unit of states
        unit_controls (list of float) : canonical unit of controls
        unit_time (float) : canonical unit of time
        p ((N,) ndarray) : inner variables for optimization
        dynamics (function) : function list, list of function of dynamics
        knot_states_smooth (list of True/False): list of states are smooth on phase knots
        cost (function) : cost function
        running_cost (function, optional) : (default = None)
        cost_derivative (function, optional) : (default = None)
        equality (function) : (default = None)
        inequality (function) : (default = None)

    """
    def _LegendreFunction(self, x, n):
        Legendre, Derivative = special.lpn(n, x)
        return Legendre[-1]

    def _LegendreDerivative(self, x, n):
        Legendre, Derivative = special.lpn(n, x)
        return Derivative[-1]

    def _nodes_LG(self, n):
        '''Return Gauss-Legendre nodes.'''
        nodes, weight = special.p_roots(n)
        return nodes

    def _weight_LG(self, n):
        '''Return Gauss-Legendre weight.'''
        nodes, weight = special.p_roots(n)
        return weight

    def _differentiation_matrix_LG(self, n):
        tau = self._nodes_LG(n)
        D = np.zeros((n, n))
        for i in range(n):
            for j in range(n):
                if i != j:
                    D[i, j] = self._LegendreDerivative(tau[i], n) \
                              / self._LegendreDerivative(tau[j], n) \
                              / (tau[i] - tau[j])
                else:
                    D[i, j] = tau[i] / (1 - tau[i]**2)
        return D

    def method_LG(self, n):
        """ Legendre-Gauss Pseudospectral method
        Gauss nodes are roots of :math:`P_n(x)`.

        Args:
            n (int) : number of nodes

        Returns:
            ndarray, ndarray, ndarray : nodes, weight, differentiation_matrix

        """
        nodes, weight = special.p_roots(n)
        D = _differentiation_matrix_LG(n)
        return nodes, weight, D

    def _nodes_LGR(self, n):
        '''Return Gauss-Radau nodes.'''
        roots, weight = special.j_roots(n-1, 0, 1)
        nodes = np.hstack((-1, roots))
        return nodes

    def _weight_LGR(self, n):
        '''Return Gauss-Legendre weight.'''
        nodes = self._nodes_LGR(n)
        w = np.zeros(0)
        for i in range(n):
            w = np.append(w, (1-nodes[i])/(n*n*self._LegendreFunction(nodes[i], n-1)**2))
        return w

    def _differentiation_matrix_LGR(self, n):
        tau = self._nodes_LGR(n)
        D = np.zeros((n, n))
        for i in range(n):
            for j in range(n):
                if i != j:
                    D[i, j] = self._LegendreFunction(tau[i], n-1) \
                              / self._LegendreFunction(tau[j], n-1) \
                              * (1 - tau[j]) / (1 - tau[i]) / (tau[i] - tau[j])
                elif i == j and i == 0:
                    D[i, j] = -(n-1)*(n+1)*0.25
                else:
                    D[i, j] = 1 / (2 * (1 - tau[i]))
        return D

    def method_LGR(self, n):
        """ Legendre-Gauss-Radau Pseudospectral method
        Gauss-Radau nodes are roots of :math:`P_n(x) + P_{n-1}(x)`.

        Args:
            n (int) : number of nodes

        Returns:
            ndarray, ndarray, ndarray : nodes, weight, differentiation_matrix

        """
        nodes = _nodes_LGR(n)
        weight = _weight_LGR(n)
        D = _differentiation_matrix_LGR(n)
        return nodes, weight, D

    def _nodes_LGL_old(self, n):
        """Return Legendre-Gauss-Lobatto nodes.
        Gauss-Lobatto nodes are roots of P'_{n-1}(x) and -1, 1.
        ref. http://keisan.casio.jp/exec/system/1360718708
        """
        x0 = np.zeros(0)
        for i in range(2, n):
            xi = (1-3.0*(n-2)/8.0/(n-1)**3)*np.cos((4.0*i-3)/(4.0*(n-1)+1)*np.pi)
            x0 = np.append(x0, xi)
        x0 = np.sort(x0)

        roots = np.zeros(0)
        for x in x0:
            optResult = optimize.root(self._LegendreDerivative, x, args=(n-1,))
            roots = np.append(roots, optResult.x)
        nodes = np.hstack((-1, roots, 1))
        return nodes

    def _nodes_LGL(self, n):
        """ Legendre-Gauss-Lobatto(LGL) points"""
        roots, weight = special.j_roots(n-2, 1, 1)
        nodes = np.hstack((-1, roots, 1))
        return nodes

    def _weight_LGL(self, n):
        """ Legendre-Gauss-Lobatto(LGL) weights."""
        nodes = self._nodes_LGL(n)
        w = np.zeros(0)
        for i in range(n):
            w = np.append(w, 2/(n*(n-1)*self._LegendreFunction(nodes[i], n-1)**2))
        return w

    def _differentiation_matrix_LGL(self, n):
        """ Legendre-Gauss-Lobatto(LGL) differentiation matrix."""
        tau = self._nodes_LGL(n)
        D = np.zeros((n, n))
        for i in range(n):
            for j in range(n):
                if i != j:
                    D[i, j] = self._LegendreFunction(tau[i], n-1) \
                              / self._LegendreFunction(tau[j], n-1) \
                              / (tau[i] - tau[j])
                elif i == j and i == 0:
                    D[i, j] = -n*(n-1)*0.25
                elif i == j and i == n-1:
                    D[i, j] = n*(n-1)*0.25
                else:
                    D[i, j] = 0.0
        return D

    def method_LGL(self, n):
        """ Legendre-Gauss-Lobatto Pseudospectral method
        Gauss-Lobatto nodes are roots of :math:`P'_{n-1}(x)` and -1, 1.

        Args:
            n (int) : number of nodes

        Returns:
            ndarray, ndarray, ndarray : nodes, weight, differentiation_matrix

        References:
            Fariba Fahroo and I. Michael Ross. "Advances in Pseudospectral Methods
            for Optimal Control", AIAA Guidance, Navigation and Control Conference
            and Exhibit, Guidance, Navigation, and Control and Co-located Conferences
            http://dx.doi.org/10.2514/6.2008-7309

        """
        nodes = _nodes_LGL(n)
        weight = _weight_LGL(n)
        D = _differentiation_matrix_LGL(n)
        return nodes, weight, D

    def _make_param_division(self, nodes, number_of_states, number_of_controls):
        prev = 0
        div = []
        for index, node in enumerate(nodes):
            num_param = number_of_states[index] + number_of_controls[index]
            temp = [i*(node) + prev for i in range(1, num_param + 1)]
            prev = temp[-1]
            div.append(temp)
        return div

    def _division_states(self, state, section):
        assert section < len(self.nodes), \
            "section argument out of own section range"
        assert state < self.number_of_states[section], \
            "states argument out of own states range"
        if (state == 0):
            if (section == 0):
                div_front = 0
            else:
                div_front = self.div[section-1][-1]
        else:
            div_front = self.div[section][state-1]
        div_back = self.div[section][state]
        return div_back, div_front

    def _division_controls(self, control, section):
        assert section < len(self.nodes), \
            "section argument out of own section range"
        assert control < self.number_of_controls[section], \
            "controls argument out of own controls range"
        div_front = self.div[section][self.number_of_states[section] + control - 1]
        div_back = self.div[section][self.number_of_states[section] + control]
        return div_back, div_front

    def states(self, state, section):
        """getter specify section states array

        Args:
            state (int) : state number
            section (int) : section number

        Returns:
            states ((N,) ndarray) :
                1-D array of state

        """
        div_back, div_front = self._division_states(state, section)
        return self.p[div_front:div_back] * self.unit_states[section][state]

    def states_all_section(self, state):
        """get states array

        Args:
            state (int) : state number

        Returns:
            states_all_section ((N,) ndarray) :
                1-D array of all section state

        """
        temp = np.zeros(0)
        for i in range(self.number_of_section):
            temp = np.concatenate([temp, self.states(state, i)])
        return temp

    def controls(self, control, section):
        """getter specify section controls array

        Args:
            control (int) : control number
            section (int) : section number

        Returns:
            controls (ndarray) :
                1-D array of controls

        """
        div_back, div_front = self._division_controls(control, section)
        return self.p[div_front:div_back] * self.unit_controls[section][control]

    def controls_all_section(self, control):
        """get controls array

        Args:
            control (int) : control number

        Returns:
            controls_all_section ((N, ) ndarray) :
                1-D array of all section control

        """
        temp = np.zeros(0)
        for i in range(self.number_of_section):
            temp = np.concatenate([temp, self.controls(control, i)])
        return temp

    def time_start(self, section):
        """ get time at section "start"

        Args:
            section (int) : section

        Returns:
            time_start (int) : time at section start

        """
        if (section == 0):
            return self.t0
        else:
            time_start_index = range(-self.number_of_section - 1, 0)
            return self.p[time_start_index[section]] * self.unit_time

    def time_final(self, section):
        """ get time at section "end"

        Args:
            section (int) : section

        Returns:
            time_final (int) : time at section end

        """
        time_final_index = range(-self.number_of_section, 0)
        return self.p[time_final_index[section]] * self.unit_time

    def time_final_all_section(self):
        """ get time at "end"

        Args:
            section (int) : section

        Returns:
            time_final_all_section (int) : time at end

        """
        tf = []
        for section in range(self.number_of_section):
            tf = tf + [self.time_final(section)]
        return tf

    def set_states(self, state, section, value):
        """set value to state at specific section

        Args:
            state (int) : state
            section (int) : section
            value (int) : value

        """
        assert len(value) == self.nodes[section], "Error: value length is NOT match nodes length"
        div_back, div_front = self._division_states(state, section)
        self.p[div_front:div_back] = value / self.unit_states[section][state]

    def set_states_all_section(self, state, value_all_section):
        """set value to state at all section

        Args:
            state (int) : state
            value_all_section (int) : value

        """
        div = 0
        for i in range(self.number_of_section):
            value = value_all_section[div:div + self.nodes[i]]
            div = div + self.nodes[i]
            self.set_states(state, i, value)

    def set_controls(self, control, section, value):
        """set value to control at all section

        Args:
            control (int) : control
            section (int) : section
            value (int) : value

        """
        assert len(value) == self.nodes[section], "Error: value length is NOT match nodes length"
        div_back, div_front = self._division_controls(control, section)
        self.p[div_front:div_back] = value / self.unit_controls[section][control]

    def set_controls_all_section(self, control, value_all_section):
        """set value to control at all section

        Args:
            control (int) : control
            value_all_section (int) : value

        """
        div = 0
        for i in range(self.number_of_section):
            value = value_all_section[div:div + self.nodes[i]]
            div = div + self.nodes[i]
            self.set_controls(control, i, value)

    def set_time_final(self, section, value):
        """ set value to final time at specific section

        Args:
            section (int) : seciton
            value (float) : value

        """
        time_final_index = range(-self.number_of_section, 0)
        self.p[time_final_index[section]] = value / self.unit_time

    def time_to_tau(self, time):
        time_init = min(time)
        time_final = max(time)
        time_center = (time_init + time_final) / 2
        temp = []
        for x in time:
            temp += [2 / (time_final - time_init) * (x - time_center)]
        return np.array(temp)

    def time_update(self):
        """ get time array after optimization

        Returns:
            time_update : (N,) ndarray
                time array

        """
        self.time = []
        t = [0] + self.time_final_all_section()
        for i in range(self.number_of_section):
            self.time.append((t[i+1] - t[i]) / 2.0 * self.tau[i]
                             + (t[i+1] + t[i]) / 2.0)
        return np.concatenate([i for i in self.time])

    def time_knots(self):
        """ get time at knot point

        Returns:
            time_knots (list) : time at knot point

        """
        return [0] + self.time_final_all_section()

    def index_states(self, state, section, index=None):
        """ get index of state at specific section

        Args:
            state (int) : state
            section (int) : section
            index (int, optional) : index

        Returns:
            index_states (int) : index of states
        """
        div_back, div_front = self._division_states(state, section)
        if (index is None):
            return div_front
        assert index < div_back - div_front, "Error, index out of range"
        if (index < 0):
            index = div_back - div_front + index
        return div_front + index

    def index_controls(self, control, section, index=None):
        div_back, div_front = self._division_controls(control, section)
        if (index is None):
            return div_front
        assert index < div_back - div_front, "Error, index out of range"
        if (index < 0):
            index = div_back - div_front + index
        return div_front + index

    def index_time_final(self, section):
        time_final_range = range(-self.number_of_section, 0)
        return self.number_of_variables + time_final_range[section]

    """
    ===========================
    UNIT SCALING ZONE
    ===========================
    """
    def set_unit_states(self, state, section, value):
        """ set a canonical unit value to the state at a specific section

        Args:
            state (int) : state
            section (int) : section
            value (float) : value

        """
        self.unit_states[section][state] = value

    def set_unit_states_all_section(self, state, value):
        """ set a canonical unit value to the state at all sections

        Args:
            state (int) : state
            value (float) : value

        """
        for i in range(self.number_of_section):
            self.set_unit_states(state, i, value)

    def set_unit_controls(self, control, section, value):
        """ set a canonical unit value to the control at a specific section

        Args:
            control (int) : control
            section (int) : section
            value (float) : value

        """
        self.unit_controls[section][control] = value

    def set_unit_controls_all_section(self, control, value):
        """ set a canonical unit value to the control at all sections

        Args:
            control (int) : control
            value (float) : value

        """
        for i in range(self.number_of_section):
            self.set_unit_controls(control, i, value)

    def set_unit_time(self, value):
        """ set a canonical unit value to the time

        Args:
            value (float) : value
        """
        self.unit_time = value
        time_init = np.array(self.time_init) / value
        self.time_init = list(time_init)
        self.time = []
        for index, node in enumerate(self.nodes):
            self.time.append((time_init[index+1] - time_init[index]) / 2.0 * self.tau[index]
                             + (time_init[index+1] + time_init[index]) / 2.0)
        self.t0 = time_init[0]
        self.time_all_section = np.concatenate([i for i in self.time])
        for section in range(self.number_of_section):
            self.set_time_final(section, time_init[section+1] * value)

    """ ==============================
    """

    def _dummy_func():
        pass

    """ ==============================
    """
    def solve(self, obj, display_func=_dummy_func, **options):
        """ solve NLP

        Args:
            obj (object instance) : instance
            display_func (function) : function to display intermediate values
            ftol (float, optional) : Precision goal for the value of f in the
                stopping criterion, (default: 1e-6)
            maxiter (int, optional) : Maximum number of iterations., (default : 25)

        Examples:
            "prob" is Problem class's instance.

            >>> prob.solve(obj, display_func, ftol=1e-12)

        """
        assert len(self.dynamics) != 0, "It must be set dynamics"
        assert self.cost is not None, "It must be set cost function"
        assert self.equality is not None, "It must be set equality function"
        assert self.inequality is not None, "It must be set inequality function"

        def equality_add(equality_func, obj):
            """ add pseudospectral method conditions to equality function.
            collocation point condition and knotting condition.
            """
            result = self.equality(self, obj)

            # collation point condition
            for i in range(self.number_of_section):
                D = self.D
                derivative = np.zeros(0)
                for j in range(self.number_of_states[i]):
                    state_temp = self.states(j, i) / self.unit_states[i][j]
                    derivative = np.hstack((derivative, D[i].dot(state_temp)))
                tix = self.time_start(i) / self.unit_time
                tfx = self.time_final(i) / self.unit_time
                dx = self.dynamics[i](self, obj, i)
                result = np.hstack((result, derivative - (tfx - tix) / 2.0 * dx))

            # knotting condition
            for knot in range(self.number_of_section - 1):
                if (self.number_of_states[knot] != self.number_of_states[knot + 1]):
                    continue  # if states are not continuous on knot, knotting condition skip
                for state in range(self.number_of_states[knot]):
                    param_prev = self.states(state, knot) / self.unit_states[knot][state]
                    param_post = self.states(state, knot + 1) / self.unit_states[knot][state]
                    if (self.knot_states_smooth[knot]):
                        result = np.hstack((result, param_prev[-1] - param_post[0]))

            return result

        def cost_add(cost_func, obj):
            """Combining nonintegrated function and integrated function.
            """
            not_integrated = self.cost(self, obj)
            if self.running_cost is None:
                return not_integrated
            integrand = self.running_cost(self, obj)
            weight = np.concatenate([i for i in self.w])
            integrated = sum(integrand * weight)
            return not_integrated + integrated

        def wrap_for_solver(func, arg0, arg1):
            def for_solver(p, arg0, arg1):
                self.p = p
                return func(arg0, arg1)
            return for_solver

        # def wrap_for_solver(func, *args):
        #     def for_solver(p, *args):
        #         self.p = p
        #         return func(*args)
        #     return for_solver

        cons = ({'type': 'eq',
                 'fun': wrap_for_solver(equality_add, self.equality, obj),
                'args': (self, obj,)},
                {'type': 'ineq',
                 'fun': wrap_for_solver(self.inequality, self, obj),
                 'args': (self, obj,)})

        if (self.cost_derivative is None):
            jac = None
        else:
            jac = wrap_for_solver(self.cost_derivative, self, obj)

        ftol = options.setdefault("ftol", 1e-6)
        maxiter = options.setdefault("maxiter", 25)

        while self.iterator < self.maxIterator:
            print("---- iteration : {0} ----".format(self.iterator+1))
            opt = optimize.minimize(wrap_for_solver(cost_add, self.cost, obj),
                                    self.p,
                                    args=(self, obj),
                                    constraints=cons,
                                    jac=jac,
                                    method='SLSQP',
                                    options={"disp": True,
                                             "maxiter": maxiter,
                                             "ftol": ftol})
            print(opt.message)
            display_func()
            print("")
            if not(opt.status):
                break
            self.iterator += 1
    """ ==============================
    """

    def __init__(self, time_init, nodes, number_of_states, number_of_controls,
                 maxIterator = 100, method="LGL"):
        assert isinstance(time_init, list), \
            "error: time_init is not list"
        assert isinstance(nodes, list), \
            "error: nodes are not list"
        assert isinstance(number_of_states, list), \
            "error: number of states are not list"
        assert isinstance(number_of_controls, list), \
            "error: number of controls are not list"
        assert len(time_init) == len(nodes) + 1, \
            "error: time_init length is not match nodes length"
        assert len(nodes) == len(number_of_states), \
            "error: nodes length is not match states length"
        assert len(nodes) == len(number_of_controls), \
            "error: nodes length is not match controls length"
        self.nodes = nodes
        self.number_of_states = number_of_states
        self.number_of_controls = number_of_controls
        self.div = self._make_param_division(nodes, number_of_states, number_of_controls)
        self.number_of_section = len(self.nodes)
        self.number_of_param = np.array(number_of_states) + np.array(number_of_controls)
        self.number_of_variables = sum(self.number_of_param * nodes) + self.number_of_section
        self.tau = []
        self.w = []
        self.D = []
        self.time = []
        for index, node in enumerate(nodes):
            self.tau.append(self._nodes_LGL(node))
            self.w.append(self._weight_LGL(node))
            self.D.append(self._differentiation_matrix_LGL(node))
            self.time.append((time_init[index+1] - time_init[index]) / 2.0 * self.tau[index]
                             + (time_init[index+1] + time_init[index]) / 2.0)
        self.maxIterator = maxIterator
        self.iterator = 0
        self.time_init = time_init
        self.t0 = time_init[0]
        self.time_all_section = np.concatenate([i for i in self.time])
        # ====
        self.unit_states = []
        self.unit_controls = []
        self.unit_time = 1.0
        for i in range(self.number_of_section):
            self.unit_states.append([1.0]*self.number_of_states[i])
            self.unit_controls.append([1.0]*self.number_of_controls[i])
        # ====
        self.p = np.zeros(self.number_of_variables, dtype=float)
        # ====
        # function
        self.dynamics = []
        self.knot_states_smooth = []
        self.cost = None
        self.running_cost = None
        self.cost_derivative = None
        self.equality = None
        self.inequality = None
        # ====
        for section in range(self.number_of_section):
            self.set_time_final(section, time_init[section+1])
            self.dynamics.append(None)
        for section in range(self.number_of_section-1):
            self.knot_states_smooth.append(True)

    def __repr__(self):
        s = "---- parameter ----" + "\n"
        s += "nodes = " + str(self.nodes) + "\n"
        s += "number of states    = " + str(self.number_of_states) + "\n"
        s += "number of controls  = " + str(self.number_of_controls) + "\n"
        s += "number of sections  = " + str(self.number_of_section) + "\n"
        s += "number of variables = " + str(self.number_of_variables) + "\n"
        s += "---- algorithm ----" + "\n"
        s += "max iteration = " + str(self.maxIterator) + "\n"
        s += "---- function  ----" + "\n"
        s += "dynamics        = " + str(self.dynamics) + "\n"
        s += "cost            = " + str(self.cost) + "\n"
        s += "cost_derivative = " + str(self.cost_derivative) + "\n"
        s += "equality        = " + str(self.equality) + "\n"
        s += "inequality      = " + str(self.inequality) + "\n"
        s += "knot_states_smooth = " + str(self.dynamics) + "\n"

        return s

    def to_csv(self, filename="OpenGoddard_output.csv", delimiter=","):
        """ output states, controls and time to csv file

        Args:
            filename (str, optional) : csv filename
            delimiter : (str, optional) : default ","

        """
        result = np.zeros(0)
        result = np.hstack((result, self.time_update()))

        header = "time, "
        for i in range(self.number_of_states[0]):
            header += "state%d, " % (i)
            result = np.vstack((result, self.states_all_section(i)))
        for i in range(self.number_of_controls[0]):
            header += "control%d, " % (i)
            result = np.vstack((result, self.controls_all_section(i)))
        np.savetxt(filename, result.T, delimiter=delimiter, header=header)
        print("Completed saving \"%s\"" % (filename))

    def plot(self, title_comment=""):
        """ plot inner variables that to be optimized

        Args:
            title_comment (str) : string for title

        """
        plt.figure()
        plt.title("OpenGoddard inner variables" + title_comment)
        plt.plot(self.p, "o")
        plt.xlabel("variables")
        plt.ylabel("value")
        for section in range(self.number_of_section):
            for line in self.div[section]:
                plt.axvline(line, color="C%d" % ((section+1) % 6), alpha=0.5)
        plt.grid()


class Guess:
    """Class for initial value guess for optimization.

    Collection of class methods
    """

    @classmethod
    def zeros(cls, time):
        """ return zeros that array size is same as time length

        Args:
            time (array_like) :

        Returns:
            (N, ) ndarray
        """
        return np.zeros(len(time))

    @classmethod
    def constant(cls, time, const):
        """ return constant values that array size is same as time length

        Args:
            time (array_like) :
            const (float) : set value

        Returns:
            (N, ) ndarray

        """
        return np.ones(len(time)) * const

    @classmethod
    def linear(cls, time, y0, yf):
        """ return linear function values that array size is same as time length

        Args:
            time (array_like) : time
            y0 (float): initial value
            yf (float): final value

        Returns:
            (N, ) ndarray

        """
        x = np.array([time[0], time[-1]])
        y = np.array([y0, yf])
        f = interpolate.interp1d(x, y)
        return f(time)

    @classmethod
    def cubic(cls, time, y0, yprime0, yf, yprimef):
        """ return cubic function values that array size is same as time length

        Args:
            time (array_like) : time
            y0 (float) : initial value
            yprime0 (float) : slope of initial value
            yf (float) : final value
            yprimef (float) : slope of final value

        Returns:
            (N, ) ndarray

        """
        y = np.array([y0, yprime0, yf, yprimef])
        t0 = time[0]
        tf = time[-1]
        A = np.array([[1, t0, t0**2, t0**3], [0, 1, 2*t0, 3*t0**2],
                      [1, tf, tf**2, tf**3], [0, 1, 2*tf, 3*tf**2]])
        invA = np.linalg.inv(A)
        C = invA.dot(y)
        ys = C[0] + C[1]*time + C[2]*time**2 + C[3]*time**3
        return ys

    @classmethod
    def plot(cls, x, y, title="", xlabel="", ylabel=""):
        """ plot wrappper

        Args:
            x (array_like) : array on the horizontal axis of the plot
            y (array_like) : array on the vertical axis of the plot
            title (str, optional) : title
            xlabel (str, optional) : xlabel
            ylabel (str, optional) : ylabel

        """
        plt.figure()
        plt.plot(x, y, "-o")
        plt.title(title)
        plt.xlabel(xlabel)
        plt.ylabel(ylabel)
        plt.grid()


class Condition(object):
    """OpenGoddard.optimize Condition class

    thin wrappper of numpy zeros and hstack

    Examples:
        for examples in equality function.

        Initial condtion : x[0] = 0.0

        Termination Condition : x[-1] = 100

        >>> result = Condition()
        >>> result.equal(x[0], 0.0)
        >>> result.equal(x[-1], 100)
        >>> return result()

        for examples in inequality function

        Inequation condtion : 0.0 <= x <= 100

        >>> result = Condition()
        >>> result.lower_bound(x, 0.0)
        >>> result.upper_bound(x, 100)
        >>> return result()

    """
    def __init__(self, length=0):
        self._condition = np.zeros(length)

    # def add(self, *args):
    #     for arg in args:
    #         self._condition = np.hstack((self._condition, arg))

    def add(self, arg, unit=1.0):
        """add condition

        Args:
            arg (array_like) : condition
        """
        self._condition = np.hstack((self._condition, arg / unit))

    def equal(self, arg1, arg2, unit=1.0):
        """add equation constraint condition in Problem equality function

        arg1 = arg2

        Args:
            arg1 (float or array_like) : right side of the equation
            arg2 (float or array_like) : left side of the equation
            unit (float, optional) : argX / unit (default : 1.0)

        Notes:
            It must be used in equality function.
        """
        arg = arg1 - arg2
        self.add(arg, unit)

    def lower_bound(self, arg1, arg2, unit=1.0):
        """add inequation constraint condition in Problem inequality function

        arg1 >= arg2

        Args:
            arg1 (array like) : arg1 is greater than or equal to arg2
            arg2 (float or array like) : arg1 is greater than or equal to arg2
            unit (float, optional) : argX / unit (default : 1.0)

        Notes:
            It must be used in inequality function.
        """
        arg = arg1 - arg2
        self.add(arg, unit)

    def upper_bound(self, arg1, arg2, unit=1.0):
        """add inequation constraint condition in Problem inequality function

        arg1 <= arg2

        Args:
            arg1 (array like) : arg1 is less than or equal to arg2
            arg2 (float or array like) : arg1 is less than or equal to arg2
            unit (float, optional) : argX / unit (default : 1.0)

        Notes:
            It must be used in inequality function.
        """
        arg = arg2 - arg1
        self.add(arg, unit)

    def change_value(self, index, value):
        self._condition[index] = value

    def __call__(self):
        return self._condition


class Dynamics(object):
    """OpenGoddard.optimize Condition class.

    thin wrapper for dynamics function.
    Behave like a dictionary type.

    Examples:
        Dynamics class must be used in dynamics function.
        It is an example of the equation of motion of thrust and free fall.
        Thrust is controllable.

        .. math::

            \dot{x} &= v

            \dot{v} &= T/m - g

        >>> def dynamics(prob, obj, section):
        >>>     x = prob.states(0, section)
        >>>     v = prob.states(1, section)
        >>>     T = prob.controls(0, section)
        >>>     g = 9.8
        >>>     m = 1.0
        >>>     dx = Dynamics(prob, section)
        >>>     dx[0] = v
        >>>     dx[1] = T / m - g
        >>>     return dx()
    """

    def __init__(self, prob, section=0):
        """ prob is instance of OpenGoddard class
        """
        self.section = section
        self.number_of_state = prob.number_of_states[section]
        self.unit_states = prob.unit_states
        self.unit_time = prob.unit_time
        for i in range(self.number_of_state):
            self.__dict__[i] = np.zeros(prob.nodes[section])

    def __getitem__(self, key):
        assert key < self.number_of_state, "Error, Dynamics key out of range"
        return self.__dict__[key]

    def __setitem__(self, key, value):
        assert key < self.number_of_state, "Error, Dynamics key out of range"
        self.__dict__[key] = value

    def __call__(self):
        dx = np.zeros(0)
        for i in range(self.number_of_state):
            temp = self.__dict__[i] * (self.unit_time / self.unit_states[self.section][i])
            dx = np.hstack((dx, temp))
        return dx


if __name__ == '__main__':
    print("==== OpenGoddard test program ====")
    plt.close("all")
    plt.ion()

    time_init = [0, 1]
    n = [10]
    num_states = [3]
    num_controls = [1]
    max_iteration = 8

    prob = Problem(time_init, n, num_states, num_controls, max_iteration)

    print("t0 = %.1f\nnodes = " % (prob.t0))
    print(n)
    print("number of states = ")
    print(num_states)
    print("number of controls = ")
    print(num_controls)
    print("tau = ")
    print(prob.tau)
    # print("D = ")
    # print(prob.D)
    print("div = ")
    print(prob.div)
    print("=====" * 10)

    time_init = [0.0, 0.10, 0.2]
    n = [20, 10]
    num_states = [3, 3]
    num_controls = [1, 1]
    max_iteration = 1

    prob = Problem(time_init, n, num_states, num_controls, max_iteration)

    print("t0 = %.1f\nnodes = " % (prob.t0))
    print(n)
    print("number of states = ")
    print(num_states)
    print("number of controls = ")
    print(num_controls)
    print("tau = ")
    print(prob.tau)
    print("time_init = ")
    print(prob.time)
    print("time_all_section = ")
    print(prob.time_all_section)
    # print("D = ")
    # print(prob.D)
    print("div = ")
    print(prob.div)
    print("=====" * 10)

    print("p = ")
    print(np.round(prob.p, 0))
    print("states #0 all section = ")
    print(prob.states_all_section(0))
    print("states #1 all section = ")
    print(prob.states_all_section(1))
    print("controls #0 section #0 = ")
    print(prob.controls(0, 0))
    print("controls #0 all section = ")
    print(prob.controls_all_section(0))
    print("time final #0 = ")
    print(prob.time_final(0))
    print("time final #1 = ")
    print(prob.time_final(1))
    print("=====" * 10)

    plt.close("all")
    tau = prob.tau
    time_init = prob.time_init
    # initial estimation
    H_init = Guess.cubic(prob.time_all_section, 1.0, 0.0, 1.01, 0.0)
    Guess.plot(prob.time_all_section, H_init, "Altitude", "time", "Altitude")
    V_init = Guess.linear(prob.time_all_section, 0.0, 0.0)
    M_init = Guess.cubic(prob.time_all_section, 1.0, -0.6, 0.6, 0.0)
    T_init1 = Guess.linear(prob.time[0], 3.5, 3.5)
    T_init2 = Guess.linear(prob.time[1], 0.0, 0.0)
    T_init = np.hstack((T_init1, T_init2))
    Guess.plot(prob.time_all_section, T_init, "Thrust Guess", "time", "Thrust")

    plt.show()

    prob.set_states_all_section(0, H_init)
    prob.set_states_all_section(1, V_init)
    prob.set_states_all_section(2, M_init)
    prob.set_controls_all_section(0, T_init)

    print("p =")
    print(np.round(prob.p, 2))
    print("=====" * 10)

    print("=====                       ====")
    print("===== Rocket Ascent Problem ====")
    print("=====                       ====")

    obj = 1.0

    def dynamics(prob, obj, section):
        h = prob.states(0, section)
        v = prob.states(1, section)
        m = prob.states(2, section)
        T = prob.controls(0, section)

        Dc = 0.5 * 620 * 1.0 / 1.0
        c = 0.5 * np.sqrt(1.0 * 1.0)
        drag = 1 * Dc * v ** 2 * np.exp(-500 * (h - 1.0) / 1.0)
        g = 1.0 * (1.0 / h)**2

        dx = np.zeros(0)
        dx0 = v
        dx1 = (T - drag) / m - g
        dx2 = - T / c
        dx = np.hstack((dx0, dx1, dx2))
        return dx

    def equality(prob, obj):
        h = prob.states_all_section(0)
        v = prob.states_all_section(1)
        m = prob.states_all_section(2)
        T = prob.controls_all_section(0)
        t0 = prob.time_start(0)
        ts = prob.time_final(0)
        tf = prob.time_final(1)

        result = Condition()

        # event condition
        result.add(h[0] - 1.0)
        result.add(v[0] - 0.0)
        result.add(m[0] - 1.0)
        result.add(v[-1] - 0.0)
        result.add(m[-1] - 0.6)

        return result()

    def inequality(prob, obj):
        h = prob.states_all_section(0)
        v = prob.states_all_section(1)
        m = prob.states_all_section(2)
        T = prob.controls_all_section(0)
        tf = prob.time_final(-1)

        result = Condition()
        # lower bounds
        result.add(h - 1.0)
        result.add(v - 0.0)
        result.add(m - 0.6)
        result.add(T - 0.0)
        result.add(tf - 0.1)
        # upper bounds
        result.add(1.0 - m)
        result.add(3.5 - T)

        return result()

    def cost(prob, obj):
        h = prob.states_all_section(0)
        return -h[-1]

    def cost_derivative(prob, obj):
        jac = Condition(prob.number_of_variables)
        index_h_end = prob.index_states(0, -1, -1)
        jac.change_value(index_h_end, -1)
        return jac()

    dx0 = dynamics(prob, obj, 0)
    dx1 = dynamics(prob, obj, 1)
    result_eq = equality(prob, obj)
    result_ineq = inequality(prob, obj)
    print("dx section #0")
    print(np.round(dx0, 2))
    print("dx section #0")
    print(np.round(dx1, 2))
    print("equality = ")
    print(np.round(result_eq, 2))
    print("inequality = ")
    print(np.round(result_ineq, 2))
    print("=====" * 10)

    prob.dynamics = [dynamics, dynamics]
    prob.knot_states_smooth = [True]
    prob.cost = cost
    prob.cost_derivative = cost_derivative
    prob.equality = equality
    prob.inequality = inequality

    def display_func():
        h = prob.states_all_section(0)
        print("max altitude: {0:.5f}".format(h[-1]))

    prob.solve(obj, display_func)

    h = prob.states_all_section(0)
    v = prob.states_all_section(1)
    m = prob.states_all_section(2)
    T = prob.controls_all_section(0)
    time = prob.time_update()

    Dc = 0.5 * 620 * 1.0 / 1.0
    drag = 1 * Dc * v ** 2 * np.exp(-500 * (h - 1.0) / 1.0)
    g = 1.0 * (1.0 / h)**2

    plt.figure()
    plt.title("Altitude profile")
    plt.plot(time, h, marker="o", label="Altitude")
    for line in prob.time_knots():
        plt.axvline(line, color="k", alpha=0.5)
    plt.grid()
    plt.xlabel("time [s]")
    plt.ylabel("Altitude [-]")

    plt.figure()
    plt.title("Velocity")
    plt.plot(time, v, marker="o", label="Velocity")
    for line in prob.time_knots():
        plt.axvline(line, color="k", alpha=0.5)
    plt.grid()
    plt.xlabel("time [s]")
    plt.ylabel("Velocity [-]")

    plt.figure()
    plt.title("Mass")
    plt.plot(time, m, marker="o", label="Mass")
    for line in prob.time_knots():
        plt.axvline(line, color="k", alpha=0.5)
    plt.grid()
    plt.xlabel("time [s]")
    plt.ylabel("Mass [-]")

    plt.figure()
    plt.title("Thrust profile")
    plt.plot(time, T, marker="o", label="Thrust")
    plt.plot(time, drag, marker="o", label="Drag")
    plt.plot(time, g, marker="o", label="Gravity")
    for line in prob.time_knots():
        plt.axvline(line, color="k", alpha=0.5)
    plt.grid()
    plt.xlabel("time [s]")
    plt.ylabel("Thrust [-]")
    plt.legend(loc="best")

    print("=====                       ====")
    print("===== Bryson-Denham Problem ====")
    print("=====                       ====")

    class BrysonDenham:
        def __init__(self):
            self.max_x = 1.0 / 9.0

    def dynamics_Bryson(prob, obj, section):
        x = prob.states(0, section)
        v = prob.states(1, section)
        u = prob.controls(0, section)

        dx = Dynamics(prob, section)
        dx[0] = v
        dx[1] = u
        return dx()

    def equality_Bryson(prob, obj):
        x = prob.states_all_section(0)
        v = prob.states_all_section(1)
        u = prob.controls_all_section(0)
        tf = prob.time_final(-1)

        result = Condition()

        result.add(x[0] - 0.0)
        result.add(v[0] - 1.0)
        result.add(x[-1] - 0.0)
        result.add(v[-1] + 1.0)
        result.add(tf - 1.0)

        return result()

    def inequality_Bryson(prob, obj):
        x = prob.states_all_section(0)
        v = prob.states_all_section(1)
        u = prob.controls_all_section(0)

        result = Condition()

        # lower bounds
        result.add(x - 0.0)

        # upper bounds
        result.add(obj.max_x - x)

        return result()

    def cost_Bryson(prob, obj):
        u = prob.controls_all_section(0)
        # integrated = 0.5 * sum(u**2 *prob.w[0])
        integrated = 0.0
        return integrated

    def running_cost_Bryson(prob, obj):
        u = prob.controls_all_section(0)
        return 0.5 * u**2

    time_init = [0, 1.0]
    nodes = [30]
    num_states = [2]
    num_controls = [1]
    max_iteration = 10

    prob = Problem(time_init, nodes, num_states, num_controls, max_iteration)
    obj = BrysonDenham()

    x_init = Guess.constant(prob.time_all_section, 0.1)
    prob.set_states_all_section(0, x_init)

    prob.dynamics = [dynamics_Bryson]
    prob.knot_states_smooth = []
    prob.cost = cost_Bryson
    prob.running_cost = running_cost_Bryson
    prob.equality = equality_Bryson
    prob.inequality = inequality_Bryson

    prob.solve(obj)
    prob.to_csv("test.csv")
    prob.plot()

    x = prob.states_all_section(0)
    v = prob.states_all_section(1)
    u = prob.controls_all_section(0)
    time = prob.time_update()

    plt.figure()
    plt.subplot(3, 1, 1)
    plt.plot(time, x, "C0-o", label="x")
    plt.ylabel("x(t)")
    plt.legend()
    plt.grid()
    plt.title("Bryson-Denham Problem")
    plt.subplot(3, 1, 2)
    plt.plot(time, v, "C1-o", label="v")
    plt.ylabel("v(t)")
    plt.legend()
    plt.grid()
    plt.subplot(3, 1, 3)
    plt.plot(time, u, "C2-o", label="u")
    plt.ylabel("u(t)")
    plt.legend()
    plt.grid()

    print("=====                   ====")
    print("===== Unit scaling Test ====")
    print("=====                   ====")
    section = 0
    x = prob.states(0, section)
    v = prob.states(1, section)
    u = prob.controls(0, section)
    dx = Dynamics(prob)
    dx[0] = v
    dx[1] = u
    dx()

    print("=====                             ====")
    print("===== Bad Brachistochrone Problem ====")
    print("=====                             ====")

    class Ball:
        def __init__(self):
            self.g = 9.8  # gravity [m/s2]
            self.l = 600000  # goal [m]
            self.h = 300000  # depth limit [m]

    def dynamics_Brachistochrone(prob, obj, section):
        x = prob.states(0, section)
        y = prob.states(1, section)
        v = prob.states(2, section)
        theta = prob.controls(0, section)

        dx = Dynamics(prob, section)
        dx[0] = v * np.sin(theta)
        dx[1] = v * np.cos(theta)
        dx[2] = obj.g * np.cos(theta)
        return dx()

    def equality_Brachistochrone(prob, obj):
        x = prob.states_all_section(0)
        y = prob.states_all_section(1)
        v = prob.states_all_section(2)
        theta = prob.controls_all_section(0)
        tf = prob.time_final(-1)

        result = Condition()
        # result.add(x[0] - 0.0)
        # result.add(y[0] - 0.0)
        # result.add(v[0] - 0.0)
        # result.add(x[-1] - obj.l)
        # result.add(y[-1] - 0.0)
        result.equal(x[0], 0.0)
        result.equal(y[0], 0.0)
        result.equal(v[0], 0.0)
        result.equal(x[-1], obj.l)
        result.equal(y[-1], 0.0)

        return result()

    def inequality_Brachistochrone(prob, obj):
        x = prob.states_all_section(0)
        y = prob.states_all_section(1)
        v = prob.states_all_section(2)
        theta = prob.controls_all_section(0)
        tf = prob.time_final(-1)

        result = Condition()
        # # lower bounds
        # result.add(tf - 500)
        # result.add(x - 0)
        # result.add(y - 0)
        # result.add(theta - 0)
        # # upper bounds
        # result.add(np.pi - theta)
        # result.add(obj.l - x)
        # result.add(700 - tf)
        # lower bounds
        result.lower_bound(tf, 500)
        result.lower_bound(x, 0)
        result.lower_bound(y, 0)
        result.lower_bound(theta, 0)
        # upper bounds
        result.upper_bound(theta, np.pi)
        result.upper_bound(x, obj.l)
        result.upper_bound(tf, 700)

        return result()

    def cost_Brachistochrone(prob, obj):
        tf = prob.time_final(-1)
        return tf

    def cost_derivative_Brachistochrone(prob, obj):
        jac = Condition(prob.number_of_variables)
        index_tf = prob.index_time_final(0)
        # index_tf = prob.index_time_final(-1)
        jac.change_value(index_tf, 1)
        return jac()

    time_init = [0.0, 700.0]
    n = [30]
    num_states = [3]
    num_controls = [1]
    max_iteration = 10
    prob = Problem(time_init, n, num_states, num_controls, max_iteration)
    obj = Ball()

    unit_x = 300000
    unit_y = 100000
    unit_time = 100
    prob.set_unit_states_all_section(0, unit_x)
    prob.set_unit_states_all_section(1, unit_y)
    prob.set_unit_states_all_section(2, unit_x / unit_time)
    prob.set_unit_controls_all_section(0, 1.0)
    prob.set_unit_time(unit_time)

    half_nodes = int(prob.nodes[0] / 2)
    theta_init = Guess.linear(prob.time_all_section, 0.0, np.pi)
    x_init = Guess.linear(prob.time_all_section, 0.0, obj.l)
    y_init0 = Guess.linear(prob.time_all_section[:half_nodes], 0, obj.h)
    y_init1 = Guess.linear(prob.time_all_section[half_nodes:], obj.h, 0)
    y_init = np.hstack((y_init0, y_init1))
    v_init0 = Guess.linear(prob.time_all_section[:half_nodes], 0, obj.h)
    v_init1 = Guess.linear(prob.time_all_section[half_nodes:], obj.h, 0)
    v_init = np.hstack((v_init0, v_init1))

    prob.set_states_all_section(0, x_init)
    prob.set_states_all_section(1, y_init)
    prob.set_controls_all_section(0, theta_init)

    prob.dynamics = [dynamics_Brachistochrone]
    prob.knot_states_smooth = []
    prob.cost = cost_Brachistochrone
    prob.cost_derivative = cost_derivative_Brachistochrone
    prob.equality = equality_Brachistochrone
    prob.inequality = inequality_Brachistochrone

    def display_func():
        tf = prob.time_final(-1)
        print("tf: {0:.5f}".format(tf))

    prob.solve(obj, display_func)

    x = prob.states_all_section(0)
    y = prob.states_all_section(1)
    v = prob.states_all_section(2)
    theta = prob.controls_all_section(0)
    time = prob.time_update()

    plt.figure()
    plt.subplot(2, 1, 1)
    plt.plot(time, x, marker="o", label="x")
    plt.plot(time, y, marker="o", label="y")
    plt.plot(time, v, marker="o", label="v")
    plt.grid()
    plt.ylabel("position [m], velocity [m/s]")
    plt.legend(loc="best")

    plt.subplot(2, 1, 2)
    plt.plot(time, theta, marker="o", label="gamma")
    plt.grid()
    plt.xlabel("time [s]")
    plt.ylabel("angle [rad]")
    plt.legend(loc="best")

    plt.figure()
    plt.plot(x, y, marker="o", label="trajectry")
    plt.grid()
    plt.xlabel("x [m]")
    plt.ylabel("y [m]")
    plt.legend(loc="best")
    plt.gca().invert_yaxis()

    prob.plot("Bad Brachistochrone")

    # plt.show()



# -*- coding: utf-8 -*-
# Copyright 2017 Interstellar Technologies Inc. All Rights Reserved.

from __future__ import print_function
import numpy as np
import matplotlib.pyplot as plt
from OpenGoddard.optimize import Problem, Guess, Condition, Dynamics

class Ball:
    def __init__(self):
        self.g = 1.0  # gravity
        self.l = 1.0  # goal
        self.h = 0.1  # limit
        self.theta0 = np.deg2rad(30)  # limit and initial angle


def dynamics(prob, obj, section):
    x = prob.states(0, section)
    y = prob.states(1, section)
    v = prob.states(2, section)
    theta = prob.controls(0, section)

    g = obj.g

    dx = Dynamics(prob, section)
    dx[0] = v * np.sin(theta)
    dx[1] = v * np.cos(theta)
    dx[2] = g * np.cos(theta)
    return dx()


def equality(prob, obj):
    x = prob.states_all_section(0)
    y = prob.states_all_section(1)
    v = prob.states_all_section(2)
    theta = prob.controls_all_section(0)
    tf = prob.time_final(-1)

    result = Condition()

    # event condition
    result.add(x[0] - 0.0)
    result.add(y[0] - 0.0)
    result.add(v[0] - 0.0)
    result.add(x[-1] - obj.l)

    return result()


def inequality(prob, obj):
    x = prob.states_all_section(0)
    y = prob.states_all_section(1)
    v = prob.states_all_section(2)
    theta = prob.controls_all_section(0)
    tf = prob.time_final(-1)

    result = Condition()

    # lower bounds
    result.add(tf - 0.1)
    result.add(y - 0)
    result.add(theta - 0)
    # result.add(x * np.tan(obj.theta0) + obj.h - y)

    # upper bounds
    # result.add(np.pi/2 - theta)

    return result()


def cost(prob, obj):
    tf = prob.time_final(-1)
    return tf


def cost_derivative(prob, obj):
    jac = Condition(prob.number_of_variables)
    # index_tf = prob.index_time_final(0)
    index_tf = prob.index_time_final(-1)
    jac.change_value(index_tf, 1)
    return jac()

# ========================
plt.close("all")
plt.ion()
# Program Starting Point
time_init = [0.0, 2.0]
n = [20]
num_states = [3]
num_controls = [1]
max_iteration = 30

flag_savefig = True

savefig_dir = "01_Brachistochrone/normal_"

# ------------------------
# set OpenGoddard class for algorithm determination
prob = Problem(time_init, n, num_states, num_controls, max_iteration)
obj = Ball()

# ========================
# Initial parameter guess
theta_init = Guess.linear(prob.time_all_section, np.deg2rad(30), np.deg2rad(30))
# Guess.plot(prob.time_all_section, theta_init, "gamma", "time", "gamma")
# if(flag_savefig):plt.savefig(savefig_dir + "guess_gamma" + savefig_add + ".png")

x_init = Guess.linear(prob.time_all_section, 0.0, obj.l)
# Guess.plot(prob.time_all_section, x_init, "x", "time", "x")
# if(flag_savefig):plt.savefig(savefig_dir + "guess_x" + savefig_add + ".png")

y_init = Guess.linear(prob.time_all_section, 0.0, obj.l / np.sqrt(3))
# Guess.plot(prob.time_all_section, theta_init, "y", "time", "y")
# if(flag_savefig):plt.savefig(savefig_dir + "guess_y" + savefig_add + ".png")

prob.set_states_all_section(0, x_init)
prob.set_states_all_section(1, y_init)
prob.set_controls_all_section(0, theta_init)

# ========================
# Main Process
# Assign problem to SQP solver
prob.dynamics = [dynamics]
prob.knot_states_smooth = []
prob.cost = cost
prob.cost_derivative = cost_derivative
prob.equality = equality
prob.inequality = inequality

prob.solve(obj)

# ========================
# Post Process
# ------------------------
# Convert parameter vector to variable
x = prob.states_all_section(0)
y = prob.states_all_section(1)
gamma = prob.controls_all_section(0)
time = prob.time_update()

# ------------------------
# Visualizetion
plt.figure()
plt.subplot(2, 1, 1)
plt.plot(time, x, marker="o", label="x")
plt.plot(time, y, marker="o", label="y")
for line in prob.time_knots():
    plt.axvline(line, color="k", alpha=0.5)
plt.grid()
plt.ylabel("velocity [m/s]")
plt.legend(loc="best")

plt.subplot(2, 1, 2)
plt.plot(time, gamma, marker="o", label="gamma")
for line in prob.time_knots():
    plt.axvline(line, color="k", alpha=0.5)
plt.grid()
plt.xlabel("time [s]")
plt.ylabel("angle [rad]")
plt.legend(loc="best")
if(flag_savefig): plt.savefig(savefig_dir + "plot" + ".png")

x_wall = np.linspace(0, obj.l)
y_wall = x_wall * np.tan(obj.theta0) + obj.h
plt.figure()
plt.plot(x, y, marker="o", label="trajectry")
# plt.plot(x_wall, y_wall, color = "k", label = "wall")
plt.axhline(0, color="k")
plt.axvline(0, color="k")
plt.axvline(obj.l, color="k")
plt.grid()
plt.xlabel("x [m]")
plt.ylabel("y [m]")
# plt.ylim([-0.02, 0.6])
plt.legend(loc="best")
plt.gca().invert_yaxis()
if(flag_savefig): plt.savefig(savefig_dir + "trajectry" + ".png")

plt.show()



# -*- coding: utf-8 -*-
# Copyright 2017 Interstellar Technologies Inc. All Rights Reserved.

from __future__ import print_function
import numpy as np
import matplotlib.pyplot as plt
from OpenGoddard.optimize import Problem, Guess, Condition, Dynamics


class Rocket:
    def __init__(self):
        self.a = 1.0  # acceleration


def dynamics(prob, obj, section):
    u = prob.states(0, section)
    v = prob.states(1, section)
    x = prob.states(2, section)
    y = prob.states(3, section)
    beta = prob.controls(0, section)

    a = obj.a

    dx = Dynamics(prob, section)
    dx[0] = a * np.cos(beta)
    dx[1] = a * np.sin(beta)
    dx[2] = u
    dx[3] = v
    return dx()


def equality(prob, obj):
    u = prob.states_all_section(0)
    v = prob.states_all_section(1)
    x = prob.states_all_section(2)
    y = prob.states_all_section(3)
    beta = prob.controls_all_section(0)
    tf = prob.time_final(-1)

    result = Condition()

    # event condition
    result.add(u[0] - 0.0)
    result.add(v[0] - 0.0)
    result.add(x[0] - 0.0)
    result.add(y[0] - 0.0)
    result.add(u[-1] - 1.0)
    result.add(v[-1] - 0.0)
    result.add(y[-1] - 1.0)

    return result()


def inequality(prob, obj):
    u = prob.states_all_section(0)
    v = prob.states_all_section(1)
    x = prob.states_all_section(2)
    y = prob.states_all_section(3)
    beta = prob.controls_all_section(0)
    tf = prob.time_final(-1)

    result = Condition()

    # lower bounds
    result.add(beta + np.pi/2)

    # upper bounds
    result.add(np.pi/2 - beta)

    return result()


def cost(prob, obj):
    tf = prob.time_final(-1)
    return tf


# ========================
plt.close("all")
# Program Starting Point
time_init = [0.0, 2.0]
n = [20]
num_states = [4]
num_controls = [1]
max_iteration = 50

flag_savefig = True
savefig_dir = "03_2d_simple_rocket/"

# ------------------------
# set OpenGoddard class for algorithm determination
prob = Problem(time_init, n, num_states, num_controls, max_iteration)
obj = Rocket()

# ========================
# Main Process
# Assign problem to SQP solver
prob.dynamics = [dynamics]
prob.knot_states_smooth = []
prob.cost = cost
prob.cost_derivative = None
prob.equality = equality
prob.inequality = inequality

prob.solve(obj)

# ========================
# Post Process
# ------------------------
# Convert parameter vector to variable
u = prob.states_all_section(0)
v = prob.states_all_section(1)
x = prob.states_all_section(2)
y = prob.states_all_section(3)
beta = prob.controls_all_section(0)
time = prob.time_update()

# ------------------------
# Visualizetion
plt.figure()
plt.subplot(3, 1, 1)
plt.plot(time, u, marker="o", label="u")
plt.plot(time, v, marker="o", label="v")
for line in prob.time_knots():
    plt.axvline(line, color="k", alpha=0.5)
plt.grid()
plt.ylabel("velocity [m/s]")
plt.legend(loc="best")

plt.subplot(3, 1, 2)
plt.plot(time, x, marker="o", label="x")
plt.plot(time, y, marker="o", label="y")
for line in prob.time_knots():
    plt.axvline(line, color="k", alpha=0.5)
plt.grid()
plt.ylabel("position [m]")
plt.legend(loc="best")

plt.subplot(3, 1, 3)
plt.plot(time, beta, marker="o", label="Mass")
for line in prob.time_knots():
    plt.axvline(line, color="k", alpha=0.5)
plt.grid()
plt.xlabel("time [s]")
plt.ylabel("angle [rad]")
plt.legend(loc="best")
if(flag_savefig): plt.savefig(savefig_dir + "plot" + ".png")

plt.figure()
plt.plot(x, y, marker="o", label="trajectry")
plt.axhline(0, color="k")
plt.axhline(1, color="k")
plt.axvline(0, color="k")
plt.grid()
plt.xlabel("x [m]")
plt.ylabel("y [m]")
plt.axis('equal')
plt.legend(loc="best")
if(flag_savefig): plt.savefig(savefig_dir + "trajectry" + ".png")

plt.show()



# -*- coding: utf-8 -*-
# Copyright 2017 Interstellar Technologies Inc. All Rights Reserved.

from __future__ import print_function
import numpy as np
import matplotlib.pyplot as plt
from OpenGoddard.optimize import Problem, Guess, Condition, Dynamics


class Rocket:
    GMe = 3.986004418 * 10**14  # Earth gravitational constant [m^3/s^2]
    Re = 6371.0 * 1000  # Earth Radius [m]
    g0 = 9.80665  # Gravitational acceleration on Earth surface [m/s^2]

    def __init__(self):
        self.Vr = np.sqrt(self.GMe / self.Re)  # m/s
        self.H0 = 10.0  # m
        self.V0 = 0.0
        self.M0 = 100000.0  # kg
        self.Mp = self.M0 * 0.99
        self.Cd = 0.6
        self.A = 4.0  # m2
        self.Isp = 300.0  # s
        self.g0 = 9.80665  # m/s2
        self.Tmax = self.M0 * self.g0 * 1.5
        self.MaxQ = 14000.0  # Pa
        self.MaxG = 8.0  # G
        self.Htarget = 400.0 * 1000  # m
        self.Rtarget = self.Re + self.Htarget  # m/s
        self.Vtarget = np.sqrt(self.GMe / self.Rtarget)  # m/s

    def air_density(self, h):
        beta = 1/8500.0  # scale factor [1/m]
        rho0 = 1.225  # kg/m3
        return rho0*np.exp(-beta*h)


def dynamics(prob, obj, section):
    R     = prob.states(0, section)
    theta = prob.states(1, section)
    Vr    = prob.states(2, section)
    Vt    = prob.states(3, section)
    m     = prob.states(4, section)
    Tr    = prob.controls(0, section)
    Tt    = prob.controls(1, section)

    rho = obj.air_density(R - obj.Re)
    Dr = 0.5 * rho * Vr * np.sqrt(Vr**2 + Vt**2) \
        * obj.Cd * obj.A  # [N]
    Dt = 0.5 * rho * Vt * np.sqrt(Vr**2 + Vt**2) \
        * obj.Cd * obj.A  # [N]
    g = obj.g0 * (obj.Re / R)**2  # [m/s2]
    g0 = obj.g0
    Isp = obj.Isp

    dx = Dynamics(prob, section)
    dx[0] = Vr
    dx[1] = Vt / R
    dx[2] = Tr / m - Dr / m - g + Vt**2 / R
    dx[3] = Tt / m - Dt / m - (Vr * Vt) / R
    dx[4] = - np.sqrt(Tr**2 + Tt**2) / g0 / Isp

    return dx()


def equality(prob, obj):
    R     = prob.states_all_section(0)
    theta = prob.states_all_section(1)
    Vr    = prob.states_all_section(2)
    Vt    = prob.states_all_section(3)
    m     = prob.states_all_section(4)
    Tr    = prob.controls_all_section(0)
    Tt    = prob.controls_all_section(1)
    tf    = prob.time_final(-1)

    result = Condition()

    # event condition
    result.add(R[0] - obj.Re, unit=prob.unit_states[0][0])
    result.add(theta[0] - 0.0, unit=prob.unit_states[0][1])
    result.add(Vr[0] - 0.0, unit=prob.unit_states[0][2])
    result.add(Vt[0] - 0.0, unit=prob.unit_states[0][3])
    result.add(m[0] - obj.M0, unit=prob.unit_states[0][4])
    result.add(R[-1] - obj.Rtarget, unit=prob.unit_states[0][1])
    result.add(Vr[-1] - 0.0, unit=prob.unit_states[0][2])
    result.add(Vt[-1] - obj.Vtarget, unit=prob.unit_states[0][3])

    return result()


def inequality(prob, obj):
    R     = prob.states_all_section(0)
    theta = prob.states_all_section(1)
    Vr    = prob.states_all_section(2)
    Vt    = prob.states_all_section(3)
    m     = prob.states_all_section(4)
    Tr    = prob.controls_all_section(0)
    Tt    = prob.controls_all_section(1)
    tf    = prob.time_final(-1)

    rho = obj.air_density(R - obj.Re)
    Dr = 0.5 * rho * Vr * np.sqrt(Vr**2 + Vt**2) \
        * obj.Cd * obj.A  # [N]
    Dt = 0.5 * rho * Vt * np.sqrt(Vr**2 + Vt**2) \
        * obj.Cd * obj.A  # [N]
    g = obj.g0 * (obj.Re / R)**2  # [m/s2]

    # dynamic pressure
    q = 0.5 * rho * (Vr**2 + Vt**2)  # [Pa]
    # accelaration
    a_r = (Tr - Dr) / m
    a_t = (Tt - Dt) / m
    a_mag = np.sqrt(a_r**2 + a_t**2)  # [m/s2]
    # Thrust
    T = np.sqrt(Tr**2 + Tt**2)

    result = Condition()

    # lower bounds
    result.add(R - obj.Re, unit=prob.unit_states[0][0])
    # result.add(Vr - 0.0, unit=prob.unit_states[0][2])
    # result.add(Vt - 0.0, unit=prob.unit_states[0][3])
    result.add(m[1:] - (obj.M0 - obj.Mp), unit=prob.unit_states[0][4])
    result.add(Tr - 0.0, unit=prob.unit_controls[0][0])
    # result.add(Tt + obj.Tmax / obj.unit_T, unit=prob.unit_controls[0][0])
    result.add(Tt - 0.0, unit=prob.unit_controls[0][0])

    # upper bounds
    result.add(obj.M0 - m, unit=prob.unit_states[0][4])
    result.add(obj.Tmax - Tr, unit=prob.unit_controls[0][0])
    result.add(obj.Tmax - Tt, unit=prob.unit_controls[0][0])
    result.add(obj.Tmax - T, unit=prob.unit_controls[0][0])
    # result.add(obj.MaxQ - q, unit = prob.unit_states[0][0])
    result.add(obj.MaxG * obj.g0 - a_mag)

    return result()


def cost(prob, obj):
    m = prob.states_all_section(4)
    # return -m[-1]
    # ==== Caution ====
    # cost function should be near 1.0
    return -m[-1] / prob.unit_states[0][4]


def cost_derivative(prob, obj):
    jac = Condition(prob.number_of_variables)
    index_m0 = prob.index_states(4, 0, 0)
    index_mf = prob.index_states(4, 0, -1)
    m = prob.states_all_section(4)
    # jac.change_value(index_m0, - m[-1] / m[0]**2)
    # jac.change_value(index_mf, - 1.0 / m[0])
    jac.change_value(index_mf, - 1.0)
    return jac()

# ========================
# plt.close("all")
plt.ion()
# Program Starting Point
time_init = [0.0, 200]
n = [30]
num_states = [5]
num_controls = [2]
max_iteration = 20

flag_savefig = True
savefig_file = "08_Rocket_Ascent_Polar/SSTO_"

# ------------------------
# set OpenGoddard class for algorithm determination
prob = Problem(time_init, n, num_states, num_controls, max_iteration)

# ------------------------
# create instance of operating object
obj = Rocket()

unit_R = obj.Re
unit_theta = 1
unit_V = np.sqrt(obj.GMe / obj.Re)
unit_m = obj.M0
unit_t = unit_R / unit_V
unit_T = unit_m * unit_R / unit_t ** 2
prob.set_unit_states_all_section(0, unit_R)
prob.set_unit_states_all_section(1, unit_theta)
prob.set_unit_states_all_section(2, unit_V)
prob.set_unit_states_all_section(3, unit_V)
prob.set_unit_states_all_section(4, unit_m)
prob.set_unit_controls_all_section(0, unit_T)
prob.set_unit_controls_all_section(1, unit_T)
prob.set_unit_time(unit_t)

# ========================
# Initial parameter guess

# altitude profile
R_init = Guess.cubic(prob.time_all_section, obj.Re, 0.0, obj.Rtarget, 0.0)
# Guess.plot(prob.time_all_section, R_init, "Altitude", "time", "Altitude")
# if(flag_savefig):plt.savefig(savefig_file + "guess_alt" + ".png")
# theta
theta_init = Guess.cubic(prob.time_all_section, 0.0, 0.0, np.deg2rad(25.0), 0.0)

# velocity
Vr_init = Guess.linear(prob.time_all_section, 0.0, 0.0)
Vt_init = Guess.linear(prob.time_all_section, 0.0, obj.Vtarget)
# Guess.plot(prob.time_all_section, V_init, "Velocity", "time", "Velocity")

# mass profile
M_init = Guess.cubic(prob.time_all_section, obj.M0, -0.6, obj.M0-obj.Mp, 0.0)
# Guess.plot(prob.time_all_section, M_init, "Mass", "time", "Mass")
# if(flag_savefig):plt.savefig(savefig_file + "guess_mass" + ".png")

# thrust profile
# T_init = Guess.zeros(prob.time_all_section)
Tr_init = Guess.cubic(prob.time_all_section, obj.Tmax/2, 0.0, 0.0, 0.0)
Tt_init = Guess.cubic(prob.time_all_section, obj.Tmax/2, 0.0, 0.0, 0.0)
# Guess.plot(prob.time_all_section, T_init, "Thrust Guess", "time", "Thrust")
# if(flag_savefig):plt.savefig(savefig_file + "guess_thrust" + ".png")

# plt.show()

# ========================
# Substitution initial value to parameter vector to be optimized
# non dimensional values (Divide by scale factor)
prob.set_states_all_section(0, R_init)
prob.set_states_all_section(1, theta_init)
prob.set_states_all_section(2, Vr_init)
prob.set_states_all_section(3, Vt_init)
prob.set_states_all_section(4, M_init)
prob.set_controls_all_section(0, Tr_init)
prob.set_controls_all_section(1, Tt_init)

# ========================
# Main Process
# Assign problem to SQP solver
prob.dynamics = [dynamics]
prob.knot_states_smooth = []
prob.cost = cost
# prob.cost_derivative = cost_derivative
prob.equality = equality
prob.inequality = inequality


def display_func():
    R = prob.states_all_section(0)
    m = prob.states_all_section(4)
    tf = prob.time_final(-1)
    print("m0          : {0:.5f}".format(m[0]))
    print("mf          : {0:.5f}".format(m[-1]))
    print("max altitude: {0:.5f}".format(R[-1]))
    print("final time  : {0:.3f}".format(tf))

prob.solve(obj, display_func, ftol=1e-8)

# ========================
# Post Process
# ------------------------
# Convert parameter vector to variable
R     = prob.states_all_section(0)
theta = prob.states_all_section(1)
Vr    = prob.states_all_section(2)
Vt    = prob.states_all_section(3)
m     = prob.states_all_section(4)
Tr    = prob.controls_all_section(0)
Tt    = prob.controls_all_section(1)
time  = prob.time_update()

# ------------------------
# Calculate necessary variables
rho = obj.air_density(R - obj.Re)
Dr  = 0.5 * rho * Vr * np.sqrt(Vr**2 + Vt**2) \
    * obj.Cd * obj.A  # [N]
Dt  = 0.5 * rho * Vt * np.sqrt(Vr**2 + Vt**2) \
    * obj.Cd * obj.A  # [N]
g   = obj.g0 * (obj.Re / R)**2  # [m/s2]

# dynamic pressure
q   = 0.5 * rho * (Vr**2 + Vt**2)  # [Pa]
# accelaration
a_r = (Tr - Dr) / m
a_t = (Tt - Dt) / m
a_mag = np.sqrt(a_r**2 + a_t**2)  # [m/s2]
# Thrust
T   = np.sqrt(Tr**2 + Tt**2)

# ------------------------
# Visualizetion
plt.figure()
plt.title("Altitude profile")
plt.plot(time, (R - obj.Re)/1000, marker="o", label="Altitude")
for line in prob.time_knots():
    plt.axvline(line, color="k", alpha=0.5)
plt.grid()
plt.xlabel("time [s]")
plt.ylabel("Altitude [km]")
if(flag_savefig): plt.savefig(savefig_file + "altitude" + ".png")

plt.figure()
plt.title("Velocity")
plt.plot(time, Vr, marker="o", label="Vr")
plt.plot(time, Vt, marker="o", label="Vt")
for line in prob.time_knots():
    plt.axvline(line, color="k", alpha=0.5)
plt.grid()
plt.xlabel("time [s]")
plt.ylabel("Velocity [m/s]")
plt.legend(loc="best")
if(flag_savefig): plt.savefig(savefig_file + "velocity" + ".png")

plt.figure()
plt.title("Mass")
plt.plot(time, m, marker="o", label="Mass")
for line in prob.time_knots():
    plt.axvline(line, color="k", alpha=0.5)
plt.grid()
plt.xlabel("time [s]")
plt.ylabel("Mass [kg]")
if(flag_savefig): plt.savefig(savefig_file + "mass" + ".png")

plt.figure()
plt.title("Acceleration")
plt.plot(time, a_r, marker="o", label="Acc r")
plt.plot(time, a_t, marker="o", label="Acc t")
plt.plot(time, a_mag, marker="o", label="Acc")
for line in prob.time_knots():
    plt.axvline(line, color="k", alpha=0.5)
plt.grid()
plt.xlabel("time [s]")
plt.ylabel("Acceleration [m/s2]")
if(flag_savefig): plt.savefig(savefig_file + "acceleration" + ".png")

plt.figure()
plt.title("Thrust profile")
plt.plot(time, Tr / 1000, marker="o", label="Tr")
plt.plot(time, Tt / 1000, marker="o", label="Tt")
plt.plot(time, T / 1000, marker="o", label="Thrust")
plt.plot(time, Dr / 1000, marker="o", label="Dr")
plt.plot(time, Dt / 1000, marker="o", label="Dt")
plt.plot(time, m * g / 1000, marker="o", label="Gravity")
for line in prob.time_knots():
    plt.axvline(line, color="k", alpha=0.5)
plt.grid()
plt.xlabel("time [s]")
plt.ylabel("Thrust [kN]")
plt.legend(loc="best")
if(flag_savefig): plt.savefig(savefig_file + "force" + ".png")

plt.show()



# -*- coding: utf-8 -*-
# Copyright 2017 Interstellar Technologies Inc. All Rights Reserved.

from __future__ import print_function
import numpy as np
import matplotlib.pyplot as plt
from OpenGoddard.optimize import Problem, Guess, Condition, Dynamics


class Rocket:
    GMe = 3.986004418 * 10**14  # Earth gravitational constant [m^3/s^2]
    Re = 6371.0 * 1000  # Earth Radius [m]
    g0 = 9.80665  # Gravitational acceleration on Earth surface [m/s^2]

    def __init__(self):
        self.M0 = 5000  # Initial total mass [kg]
        self.Mc = 0.4  # Initial Propellant mass over total mass
        self.Cd = 0.2  # Drag Coefficient [-]
        self.area = 10  # area [m2]
        self.Isp = 300.0  # Isp [s]
        self.max_thrust = 2  # maximum thrust to initial weight ratio

    def air_density(self, h):
        beta = 1/8500.0  # scale factor [1/m]
        rho0 = 1.225  # kg/m3
        return rho0*np.exp(-beta*h)


def dynamics(prob, obj, section):
    R = prob.states(0, section)
    v = prob.states(1, section)
    m = prob.states(2, section)
    T = prob.controls(0, section)

    rho = obj.air_density(R - obj.Re)
    drag = 0.5 * rho * v ** 2 * obj.Cd * obj.area
    g = obj.GMe / R**2
    g0 = obj.g0
    Isp = obj.Isp

    dx = Dynamics(prob, section)
    dx[0] = v
    dx[1] = (T - drag) / m - g
    dx[2] = - T / g0 / Isp
    return dx()


def equality(prob, obj):
    R = prob.states_all_section(0)
    v = prob.states_all_section(1)
    m = prob.states_all_section(2)
    T = prob.controls_all_section(0)
    tf = prob.time_final(-1)

    result = Condition()

    # event condition
    result.add(R[0] - obj.Re)
    result.add(v[0] - 0.0)
    result.add(m[0] - obj.M0)
    result.add(v[-1] - 0.0)
    result.add(m[-1] - obj.M0 * obj.Mc)

    return result()


def inequality(prob, obj):
    R = prob.states_all_section(0)
    v = prob.states_all_section(1)
    m = prob.states_all_section(2)
    T = prob.controls_all_section(0)
    tf = prob.time_final(-1)

    result = Condition()
    # lower bounds
    result.add(R - obj.Re)
    result.add(v - 0.0)
    result.add(m - obj.M0 * obj.Mc)
    result.add(T - 0.0)
    result.add(tf - 10)
    # upper bounds
    result.add(obj.M0 - m)
    result.add(obj.max_thrust * obj.M0 * obj.g0 - T)

    return result()


def cost(prob, obj):
    R = prob.states_all_section(0)
    return -R[-1] / obj.Re


def cost_derivative(prob, obj):
    jac = Condition(prob.number_of_variables)
    index_R_end = prob.index_states(0, 0, -1)
    jac.change_value(index_R_end, -1)
    return jac()

# ========================
plt.close("all")
plt.ion()
# Program Starting Point
time_init = [0.0, 600]
n = [50]
num_states = [3]
num_controls = [1]
max_iteration = 20

flag_savefig = True
savefig_file = "06_Rocket_Ascent/Single_"

# ------------------------
# set OpenGoddard class for algorithm determination
prob = Problem(time_init, n, num_states, num_controls, max_iteration)

# ------------------------
# create instance of operating object
obj = Rocket()

# ------------------------
# set designer unit
unit_R = obj.Re
unit_v = np.sqrt(obj.GMe / obj.Re)
unit_m = obj.M0
unit_t = unit_R / unit_v
unit_T = unit_m * unit_R / unit_t**2
prob.set_unit_states_all_section(0, unit_R)
prob.set_unit_states_all_section(1, unit_v)
prob.set_unit_states_all_section(2, unit_m)
prob.set_unit_controls_all_section(0, unit_T)
prob.set_unit_time(unit_t)

# ========================
# Initial parameter guess

# altitude profile
R_init = Guess.cubic(prob.time_all_section, obj.Re, 0.0, obj.Re+50*1000, 0.0)
# Guess.plot(prob.time_all_section, R_init, "Altitude", "time", "Altitude")
# if(flag_savefig):plt.savefig(savefig_file + "guess_alt" + ".png")

# velocity
V_init = Guess.linear(prob.time_all_section, 0.0, 0.0)
# Guess.plot(prob.time_all_section, V_init, "Velocity", "time", "Velocity")

# mass profile
M_init = Guess.cubic(prob.time_all_section, obj.M0, -0.6, obj.M0*obj.Mc, 0.0)
# Guess.plot(prob.time_all_section, M_init, "Mass", "time", "Mass")
# if(flag_savefig):plt.savefig(savefig_file + "guess_mass" + ".png")

# thrust profile
T_init = Guess.cubic(prob.time_all_section, obj.max_thrust * obj.M0 * obj.g0, 0.0, 0.0, 0.0)
# Guess.plot(prob.time_all_section, T_init, "Thrust Guess", "time", "Thrust")
# if(flag_savefig):plt.savefig(savefig_file + "guess_mass" + ".png")

plt.show()

# ========================
# Substitution initial value to parameter vector to be optimized
prob.set_states_all_section(0, R_init)
prob.set_states_all_section(1, V_init)
prob.set_states_all_section(2, M_init)
prob.set_controls_all_section(0, T_init)

# ========================
# Main Process
# Assign problem to SQP solver
prob.dynamics = [dynamics]
prob.knot_states_smooth = []
prob.cost = cost
# prob.cost_derivative = cost_derivative
prob.equality = equality
prob.inequality = inequality


def display_func():
    R = prob.states_all_section(0)
    print("max altitude: {0:.5f}".format(R[-1] - obj.Re))

prob.solve(obj, display_func, ftol=1e-12)

# ========================
# Post Process
# ------------------------
# Convert parameter vector to variable
R = prob.states_all_section(0)
v = prob.states_all_section(1)
m = prob.states_all_section(2)
T = prob.controls_all_section(0)
time = prob.time_update()

# ------------------------
# Calculate necessary variables
rho = obj.air_density(R - obj.Re)
drag = 0.5 * rho * v ** 2 * obj.Cd * obj.area
g = obj.GMe / R**2

# ------------------------
# Visualizetion
plt.figure()
plt.title("Altitude profile")
plt.plot(time, (R - obj.Re)/1000, marker="o", label="Altitude")
for line in prob.time_knots():
    plt.axvline(line, color="k", alpha=0.5)
plt.grid()
plt.xlabel("time [s]")
plt.ylabel("Altitude [km]")
if(flag_savefig): plt.savefig(savefig_file + "altitude" + ".png")

plt.figure()
plt.title("Velocity")
plt.plot(time, v, marker="o", label="Velocity")
for line in prob.time_knots():
    plt.axvline(line, color="k", alpha=0.5)
plt.grid()
plt.xlabel("time [s]")
plt.ylabel("Velocity [m/s]")
if(flag_savefig): plt.savefig(savefig_file + "velocity" + ".png")

plt.figure()
plt.title("Mass")
plt.plot(time, m, marker="o", label="Mass")
for line in prob.time_knots():
    plt.axvline(line, color="k", alpha=0.5)
plt.grid()
plt.xlabel("time [s]")
plt.ylabel("Mass [kg]")
if(flag_savefig): plt.savefig(savefig_file + "mass" + ".png")

plt.figure()
plt.title("Thrust profile")
plt.plot(time, T / 1000, marker="o", label="Thrust")
plt.plot(time, drag / 1000, marker="o", label="Drag")
plt.plot(time, m * g / 1000, marker="o", label="Gravity")
for line in prob.time_knots():
    plt.axvline(line, color="k", alpha=0.5)
plt.grid()
plt.xlabel("time [s]")
plt.ylabel("Thrust [kN]")
plt.legend(loc="best")
if(flag_savefig): plt.savefig(savefig_file + "force" + ".png")

plt.show()



# -*- coding: utf-8 -*-
# Copyright 2017 Interstellar Technologies Inc. All Rights Reserved.

from __future__ import print_function
import numpy as np
import matplotlib.pyplot as plt
from OpenGoddard.optimize import Problem, Guess, Condition, Dynamics


class Rocket:
    g0 = 1.0  # Gravity at surface [-]

    def __init__(self):
        self.H0 = 1.0  # Initial height
        self.V0 = 0.0  # Initial velocity
        self.M0 = 1.0  # Initial mass
        self.Tc = 3.5  # Use for thrust
        self.Hc = 500  # Use for drag
        self.Vc = 620  # Use for drag
        self.Mc = 0.6  # Fraction of initial mass left at end
        self.c = 0.5 * np.sqrt(self.g0*self.H0)  # Thrust-to-fuel mass
        self.Mf = self.Mc * self.M0               # Final mass
        self.Dc = 0.5 * self.Vc * self.M0 / self.g0  # Drag scaling
        self.T_max = self.Tc * self.g0 * self.M0     # Maximum thrust


def dynamics(prob, obj, section):
    h = prob.states(0, section)
    v = prob.states(1, section)
    m = prob.states(2, section)
    T = prob.controls(0, section)

    Dc = obj.Dc
    c = obj.c
    drag = 1 * Dc * v ** 2 * np.exp(-obj.Hc * (h - obj.H0) / obj.H0)
    g = obj.g0 * (obj.H0 / h)**2

    dx = Dynamics(prob, section)
    dx[0] = v
    dx[1] = (T - drag) / m - g
    dx[2] = - T / c
    return dx()


def equality(prob, obj):
    h = prob.states_all_section(0)
    v = prob.states_all_section(1)
    m = prob.states_all_section(2)
    T = prob.controls_all_section(0)
    tf = prob.time_final(-1)

    result = Condition()

    # event condition
    result.add(h[0] - obj.H0)
    result.add(v[0] - obj.V0)
    result.add(m[0] - obj.M0)
    result.add(v[-1] - 0.0)
    result.add(m[-1] - obj.Mf)

    return result()


def inequality(prob, obj):
    h = prob.states_all_section(0)
    v = prob.states_all_section(1)
    m = prob.states_all_section(2)
    T = prob.controls_all_section(0)
    tf = prob.time_final(-1)

    result = Condition()
    # lower bounds
    result.add(h - obj.H0)
    result.add(v - 0.0)
    result.add(m - obj.Mf)
    result.add(T - 0.0)
    result.add(tf - 0.1)
    # upper bounds
    result.add(obj.M0 - m)
    result.add(obj.T_max - T)

    return result()


def cost(prob, obj):
    h = prob.states_all_section(0)
    return -h[-1]


# ========================
plt.close("all")
# Program Starting Point
time_init = [0.0, 0.3]
n = [50]
num_states = [3]
num_controls = [1]
max_iteration = 30

flag_savefig = True
savefig_file = "04_Goddard/04_0knot_"

# ------------------------
# set OpenGoddard class for algorithm determination
prob = Problem(time_init, n, num_states, num_controls, max_iteration)

# ------------------------
# create instance of operating object
# Nondimensionalization of parameters
obj = Rocket()

# ========================
# Initial parameter guess

# altitude profile
H_init = Guess.cubic(prob.time_all_section, 1.0, 0.0, 1.010, 0.0)
# Guess.plot(prob.time_all_section, H_init, "Altitude", "time", "Altitude")
# if(flag_savefig):plt.savefig(savefig_file + "guess_alt" + ".png")

# velocity
V_init = Guess.linear(prob.time_all_section, 0.0, 0.0)
# Guess.plot(prob.time_all_section, V_init, "Velocity", "time", "Velocity")

# mass profile
M_init = Guess.cubic(prob.time_all_section, 1.0, -0.6, 0.6, 0.0)
# Guess.plot(prob.time_all_section, M_init, "Mass", "time", "Mass")
# if(flag_savefig):plt.savefig(savefig_file + "guess_mass" + ".png")

# thrust profile
T_init = Guess.cubic(prob.time_all_section, 3.5, 0.0, 0.0, 0.0)
# Guess.plot(prob.time_all_section, T_init, "Thrust Guess", "time", "Thrust")
# if(flag_savefig):plt.savefig(savefig_file + "guess_thrust" + ".png")

plt.show()

# ========================
# Substitution initial value to parameter vector to be optimized
prob.set_states_all_section(0, H_init)
prob.set_states_all_section(1, V_init)
prob.set_states_all_section(2, M_init)
prob.set_controls_all_section(0, T_init)

# ========================
# Main Process
# Assign problem to SQP solver
prob.dynamics = [dynamics]
prob.knot_states_smooth = []
prob.cost = cost
prob.cost_derivative = None
prob.equality = equality
prob.inequality = inequality


def display_func():
    h = prob.states_all_section(0)
    print("max altitude: {0:.5f}".format(h[-1]))

prob.solve(obj, display_func, ftol=1e-10)

# ========================
# Post Process
# ------------------------
# Convert parameter vector to variable
h = prob.states_all_section(0)
v = prob.states_all_section(1)
m = prob.states_all_section(2)
T = prob.controls_all_section(0)
time = prob.time_update()

# ------------------------
# Calculate necessary variables
Dc = 0.5 * 620 * 1.0 / 1.0
drag = 1 * Dc * v ** 2 * np.exp(-500 * (h - 1.0) / 1.0)
g = 1.0 * (1.0 / h)**2

# ------------------------
# Visualizetion
plt.figure()
plt.title("Altitude profile")
plt.plot(time, h, marker="o", label="Altitude")
for line in prob.time_knots():
    plt.axvline(line, color="k", alpha=0.5)
plt.grid()
plt.xlabel("time [s]")
plt.ylabel("Altitude [-]")
if(flag_savefig): plt.savefig(savefig_file + "altitude" + ".png")

plt.figure()
plt.title("Velocity")
plt.plot(time, v, marker="o", label="Velocity")
for line in prob.time_knots():
    plt.axvline(line, color="k", alpha=0.5)
plt.grid()
plt.xlabel("time [s]")
plt.ylabel("Velocity [-]")
if(flag_savefig): plt.savefig(savefig_file + "velocity" + ".png")

plt.figure()
plt.title("Mass")
plt.plot(time, m, marker="o", label="Mass")
for line in prob.time_knots():
    plt.axvline(line, color="k", alpha=0.5)
plt.grid()
plt.xlabel("time [s]")
plt.ylabel("Mass [-]")
if(flag_savefig): plt.savefig(savefig_file + "mass" + ".png")

plt.figure()
plt.title("Thrust profile")
plt.plot(time, T, marker="o", label="Thrust")
plt.plot(time, drag, marker="o", label="Drag")
plt.plot(time, g, marker="o", label="Gravity")
for line in prob.time_knots():
    plt.axvline(line, color="k", alpha=0.5)
plt.grid()
plt.xlabel("time [s]")
plt.ylabel("Thrust [-]")
plt.legend(loc="best")
if(flag_savefig): plt.savefig(savefig_file + "force" + ".png")

plt.show()
